@@ -1,10 +1,8 @@
-from __future__ import unicode_literals
-
 from operator import attrgetter
 
-from django.apps.registry import Apps
 from django.db import models
-from django.test import TestCase
+from django.test import SimpleTestCase, TestCase
+from django.test.utils import isolate_apps
 
 from .base_tests import BaseOrderWithRespectToTests
 from .models import Answer, Dimension, Entity, Post, Question
@@ -16,24 +14,20 @@ class OrderWithRespectToBaseTests(BaseOrderWithRespectToTests, TestCase):
     Question = Question
 
 
-class OrderWithRespectToTests(TestCase):
-
+class OrderWithRespectToTests(SimpleTestCase):
+    @isolate_apps("order_with_respect_to")
     def test_duplicate_order_field(self):
-        test_apps = Apps(['order_with_respect_to'])
-
         class Bar(models.Model):
             class Meta:
-                apps = test_apps
-                app_label = 'order_with_respect_to'
+                app_label = "order_with_respect_to"
 
         class Foo(models.Model):
             bar = models.ForeignKey(Bar, models.CASCADE)
             order = models.OrderWrt()
 
             class Meta:
-                order_with_respect_to = 'bar'
-                apps = test_apps
-                app_label = 'order_with_respect_to'
+                order_with_respect_to = "bar"
+                app_label = "order_with_respect_to"
 
         count = 0
         for field in Foo._meta.local_fields:
@@ -50,4 +44,6 @@ class TestOrderWithRespectToOneToOnePK(TestCase):
         c1 = d.component_set.create()
         c2 = d.component_set.create()
         d.set_component_order([c1.id, c2.id])
-        self.assertQuerysetEqual(d.component_set.all(), [c1.id, c2.id], attrgetter('pk'))
+        self.assertQuerysetEqual(
+            d.component_set.all(), [c1.id, c2.id], attrgetter("pk")
+        )
